
//
//
// Project       : Vorpal-Mail
//
// Source        : server.cc
//
// Creator       : Blane Bramble                    Date          : 23rd October 2000
// Amended       : Blane Bramble                    Date          : 7th December 2000
//
// License       : Copyright Virtua-Web (GPL to be used once version 1.0 is achieved)
//
// SMTP daemon server, utilising SMTP_SESSION class (session.h, session.cc). Provides
// an RFC 821 compliant SMTP server daemon.
//
//
// Changelog:
//
// 5th December 2000
// Additional code to allow Vorpal Mail to function as a replacement for sendmail
// from the command line (various programs seem to call sendmail directly to
// generate emails).
//
// 7th December 2000
// Limit specifying a FROM user for sendmail mode to root. Otherwise use
// the uid and gecos information.
//

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <getopt.h>
#include <dirent.h>
#include <fcntl.h>
#include <syslog.h>
#include <pwd.h>

#include <netdb.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include "mail.h"
#include "array.h"
#include "session.h"
#include "list.h"
#include "virtual.h"
#include "smtp.h"

SMTP_SESSION *session = NULL, *client = NULL;

VIRTUAL_DOMAIN *domain;

LIST *smtplist = NULL;

int run = 0, debug = 0, interval = 0, reloaded = 0;

int control_socket = -1;
long retrytime = 300;				// 5 minute gap before a msg is retried
int cached = 0;

static struct option longoptions[] = {
	{ "list", 0, NULL, 'l', },
	{ "help", 0, NULL, 'h', },
	{ "daemon", 0, NULL, 'd', },
	{ "queue", 0, NULL, 'q', },
	{ "dump", 0, NULL, 'm', },
	{ "debug", 2, NULL, 'D', },
	{ "retry", 1, NULL, 'r', },
	{ NULL, 0, NULL, 0, },
	};

char *personality[] = {
	"mailq",
	"sendmail",
	"send-mail",
	NULL,
	};

enum	{
	PERS_DEFAULT = -1,
	PERS_MAILQ = 0,
	PERS_SENDMAIL,
	PERS_SEND_MAIL,
	PERS_LAST
	};

void signalman(int sig)
{
int status, pid;
switch(sig)
	{
	case SIGINT:
		syslog(LOG_MAIL + LOG_NOTICE, "SIGINT received. Exiting.\n");
		run = 0;
		break;
	case SIGTERM:
		syslog(LOG_MAIL + LOG_NOTICE, "SIGTERM received. Exiting.\n");
		run = 0;
		break;
	case SIGHUP:
		{
		LISTITEM *item;
		delete domain;
		domain = new VIRTUAL_DOMAIN();
		domain->load(CONFIG);
		if (smtplist != NULL)
			{
			item = smtplist->gethead();
			while(item != NULL)
				{
				((SMTP_SESSION *)(item->getdata()))->setdomain(domain);
				item = item->getnext();
				}
			}
		syslog(LOG_MAIL + LOG_NOTICE, "SIGHUP received. Configuration reloaded.\n");
		}
		break;
	case SIGALRM:
		interval++;
		break;
	}
signal(sig, signalman);
}

int makedaemon(void)
{
switch(fork())
	{
	case -1:				// failed, log msg and exit
		syslog(LOG_MAIL + LOG_EMERG, "** Could not fork process. Cannot run as daemon");
		return 0;
	case 0:					// child, so do more stuff
		if (setsid() == -1)
			{
			syslog(LOG_MAIL + LOG_EMERG, "** Could not setsid. Cannot run as daemon");
			return 0;
			}
		switch(fork())
			{
			case -1:		// failed, log and exit
				syslog(LOG_MAIL + LOG_EMERG, "** Could not fork process (2). Cannot run as daemon");
				return 0;
			case 0:			// grandchild, so do more stuff
				chdir("/");	// change to root
				umask(0);	// haha! he would have got away with it to, if it wasn't for us darn kids
				freopen("/dev/null", "r", stdin);
				freopen("/dev/console", "w", stdout);
				freopen("/dev/console", "w", stderr);
				return 1;	// All worship me, for I am an all powerful Daemon
			default:		// child - so can exit
				return 0;
			}
		break;		
	default:				// parent - so can exit
		return 0;
	}
}

int clientinitialise(char *who)
{
long flags;
struct sockaddr_in sad;

struct hostent *host;

if (control_socket == -1)
	{

	if ((control_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		{
		syslog(LOG_MAIL + LOG_EMERG, "** Could not create socket.\n");
		return 0;
		}


	if ((host = gethostbyname(who)) == NULL)
		{
		if (debug)
			syslog(LOG_MAIL + LOG_INFO, "** Could not get relay by name");
		control_socket = -1;
		printf("Can't get relay by name\n");
		return 0;
		}

	sad.sin_family = AF_INET;
	//sad.sin_addr.s_addr = ((struct sockaddr_in *)(host->h_addr))->sin_addr.s_addr;
	sad.sin_addr.s_addr = *(unsigned long *)(host->h_addr);
	sad.sin_port = htons(25);

	if (connect(control_socket, (sockaddr *)&sad, sizeof(sad)) < 0)
		{
		syslog(LOG_MAIL + LOG_EMERG, "** Could not connect socket.\n");
		printf("Can't connect to relay\n");
		return 0;
		}
	flags = 1;
	ioctl(control_socket, FIONBIO, &flags);
	signal(SIGTERM, signalman);
	signal(SIGHUP, signalman);
	signal(SIGINT, signalman);

	client = new SMTP_SESSION(control_socket, &sad);

	cached = 0;
	}
else
	cached = 1;

client->setmode(SMTP_MODE_CLIENT);
client->setdebug(debug);
client->setdomain(domain);

return 1;
}

int initialise(void)
{
long flags;
struct sockaddr_in sad;
struct itimerval timer;

if ((control_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0)	/* create a socket, or fail initialisation */
	return 0;

flags = 1;					/* Set non-blocking I/O mode (flags = 1) */
ioctl(control_socket, FIONBIO, &flags);

sad.sin_family = AF_INET;
sad.sin_addr.s_addr = INADDR_ANY;
sad.sin_port = htons(domain->getport());			/* request a contol socket listening on any addr, port as specified */

if (bind(control_socket, (sockaddr *)&sad, sizeof(sad)) < 0)
	return 0;				/* bind the address to our socket, or fail initialisation */

if (listen(control_socket, 5) != 0)
	return 0;				/* start listening (or fail...) */

signal(SIGTERM, signalman);
signal(SIGHUP, signalman);
signal(SIGINT, signalman);
signal(SIGALRM, signalman);

timer.it_value.tv_sec = 10;
timer.it_interval.tv_sec = 10;

if (setitimer(ITIMER_REAL, &timer, NULL) != 0)
	return 0;

return 1;					/* hey, it all worked! */
}

int clientrun(int cache)
{
struct timeval tv;

run = 1;

while(run)
	{
	fd_set fdr, fdw;
	FD_ZERO(&fdr);
	FD_ZERO(&fdw);
	FD_SET(control_socket, &fdr);
	if (client->getstatus() == SMTP_STATUS_DATASEND || client->pending())
		FD_SET(control_socket, &fdw);
	tv.tv_sec = 10; tv.tv_usec = 0;

//	if (cache)						// no connect msg as re-using socket, so spoof a 250
//		client->process("250 cached connection\r\n");
	if (cache)
		client->sendreset();

	cache = 0;

	if (select(FD_SETSIZE, &fdr, &fdw, NULL, &tv) > 0)
		{
		if (FD_ISSET(control_socket, &fdr))
			{
			char buf[SMTP_MAX_LINE];
			int t;
			if ((t=recv(control_socket, buf, SMTP_MAX_LINE, 0))<1)
				{
				syslog(LOG_MAIL + LOG_INFO, "Closed connection removed.\n");
				run = 0;
				}
			else
				{
/* 0 terminate the input string */
				*(buf + t) = 0;
				client->process(buf);
				switch(client->getstatus())
					{
					case SMTP_STATUS_ABORT:
						run = 0;
						break;
					case SMTP_STATUS_COMPLETE:
						run = 0;
						break;
					}
				}
			}
		if (control_socket != -1 && FD_ISSET(control_socket, &fdw))
			{
			if (client->getstatus() == SMTP_STATUS_DATASEND)
				{
				client->process("\r\n");
				}
			client->dosend();
			}
		}
	}
}

int serverrun(void)
{
int n;
struct timeval tv;
LISTITEM *item;
int pid, status;

if (!initialise())
	{
	printf("Could not initialise.\n");
	return 1;
	}

if ((smtplist = new LIST) == NULL)
	return 1;

if (!makedaemon())
	return 0;

syslog(LOG_MAIL + LOG_NOTICE, PROGRAM " " VERSION " daemon started");

run = 1;

while(run)
	{
	fd_set fdr,fdw;
	FD_ZERO(&fdr);
	FD_ZERO(&fdw);
	FD_SET(control_socket,&fdr);

	item = smtplist->gethead();
	while(item != NULL)
		{
		int sock;

		session = (SMTP_SESSION *)(item->getdata());

		if ((sock = session->getsocket()) != -1)
			{
			FD_SET(sock, &fdr);
			if (session->pending())
				FD_SET(sock, &fdw);
			if (interval)
				{
				session->pulse();
				interval--;
				}
			}
		item = item->getnext();
		}

	if ((pid = waitpid(-1, &status, WNOHANG)) > 0 && debug)
		syslog(LOG_MAIL + LOG_DEBUG, "Child process %d has exitted", pid);

	tv.tv_sec=10; tv.tv_usec=0;
	if (select(FD_SETSIZE,&fdr,&fdw,NULL,&tv) > 0)
		{
/* New Terminal Connected */
		if (FD_ISSET(control_socket, &fdr))
			{
/* New Socket, Addrlength, Line */
			int ns,al,li;
			struct sockaddr_in sa;
			al = sizeof(sa);
			if ((ns=accept(control_socket, (sockaddr*)&sa,(socklen_t*)&al))>-1)
				{
				long fl=1;
				SMTP_SESSION *sess;

				ioctl(ns,FIONBIO,&fl);

				if ((sess = new SMTP_SESSION(ns, &sa)) == NULL)
					{
					shutdown(ns, 2);
					close(ns);
					}
				else
					{
					smtplist->add(NULL, sess);
					}
				sess->setdebug(debug);
				sess->setdomain(domain);
				sess->setmode(SMTP_MODE_SERVER);
				sess->serverstatus(220);
				}
			}
		item = smtplist->gethead();
		while(item != NULL)
			{
			int sock;
			LISTITEM *next;

			next = item->getnext();
			session = ((SMTP_SESSION *)(item->getdata()));
			if ((sock = session->getsocket())  != -1 && FD_ISSET(sock, &fdr))
				{
				char buf[SMTP_MAX_LINE];
				int t;
				if ((t=recv(sock, buf, SMTP_MAX_LINE, 0))<1)
					{
					delete (SMTP_SESSION *)smtplist->del(item);
					syslog(LOG_MAIL + LOG_INFO, "Closed connection removed.\n");
					}
				else
					{
/* 0 terminate the input string */
					*(buf + t) = 0;
					session->process(buf);
					if (session->getstatus() == SMTP_STATUS_COMPLETE)
						{
						delete (SMTP_SESSION *)smtplist->del(item);
						}
/* Need to add code to handle received input */

					}
				}
			if (sock != -1 && FD_ISSET(sock, &fdw))
				{	
				session->dosend();
				}

			item = next;
			}
		}
	}

item = smtplist->gethead();
while(item != NULL)
	{
	LISTITEM *next;
	next = item->getnext();
	delete (SMTP_SESSION *)smtplist->del(item);
	syslog(LOG_MAIL + LOG_INFO, "Zombie link removed.\n");
	item = next;
	}

syslog(LOG_MAIL + LOG_NOTICE, PROGRAM " " VERSION " daemon shutdown");

return 0;
}


void help(void)
{
printf(COPYRIGHT "\n");
printf("\n");
printf("Arguments:\n");
printf("-h, --help            Displays this help\n");
printf("-l, --list            List the outgoing queue\n");
printf("-d, --daemon          Run as a mail server (daemon)\n");
printf("-q, --queue           Process and send outgoing mail\n");
printf("-D<n>, --debug=<n>    Sets logging level (1 if no number specified)\n");
printf("-r, --retry <t>       Set minimum time before resending a message\n");
printf("\n");
printf("This program is released under the GNU General Public License (GPL) Version 2.\n");
printf("Please see the supplied file gpl.txt, or the GNU Web-site at www.gnu.org\n");
}

int processitem(char *id)
{
char filename[PATH_MAX];
char line[SMTP_MAX_LINE];
FILE *header;
int mature = 1;

if (!clientinitialise(domain->getsmtp()))
	return 0;

snprintf(filename, PATH_MAX-1, "%s/H%s", domain->getqout(), id);

if ((header = fopen(filename, "r")) != NULL)
	{
	int first = 1;
	int fd;
	int locked = 1;
	char *dfile;
	struct flock flk;


	snprintf(filename, PATH_MAX-1, "%s/R%s", domain->getqout(), id);
	if ((fd = open(filename, O_RDONLY)) != -1)
		{
		fcntl(fd, F_GETLK, &flk);
		if (flk.l_type == F_UNLCK)
			locked = 0;
		close(fd);
		}

	if (locked)
		{
		printf("Locked\n");
		fclose(header);
		return 0;
		}

				
	while(fgets(line, SMTP_MAX_LINE, header) != NULL)
		{
		char *ptr;
		if ((ptr = strchr(line, '\n')) != NULL)
			*ptr = 0;
		switch(*line)
			{
			case 'F':
				client->setfrom(line+1);
				break;
			case 'T':
				client->addto(line+1);
				break;
			case 'Q':
				{
				time_t now,t;
				now = time(NULL);
				t = strtol(line+1, NULL, 10);
				client->setqtime(t);
				}
				break;
			case 'L':
				{
				time_t now, t;
				now = time(NULL);
				t = strtol(line+1, NULL, 10);
				if ((now - t) < retrytime)
					mature = 0;
				client->setltime(t);
				}
				break;
			}
		}
	fclose(header);

	if (mature)
		{	
		client->openfiles(id);

		printf("Sending.\n");
	
		clientrun(cached);
		}

	client->reset();

	}
else
	printf("%s Locked\n", filename);
return 0;	
}

int queuerun(void)
{
char cwd[PATH_MAX];
char dir[PATH_MAX];
char line[SMTP_MAX_LINE];
DIR *queue;
int notprocessed = 0;

getcwd(cwd, PATH_MAX);

chdir(domain->getqout());
getcwd(dir, PATH_MAX);

if ((queue = opendir(dir)) == NULL)
	printf("Could not open queue directory\n");
else
	{
	struct dirent *file;
	while((file = readdir(queue)) != NULL)
		{
		FILE *header;
		if (*(file->d_name) == 'H')
			{
			printf("Processing SMTP id %s\n", file->d_name + 1);
			notprocessed += processitem(file->d_name+1);
			}
		}
	closedir(queue);
	}
chdir(cwd);
if (client != NULL)
	client->sendquit();
return notprocessed;
}

void listqueue(void)
{
char cwd[PATH_MAX];
char dir[PATH_MAX];
char line[SMTP_MAX_LINE];
DIR *queue;
struct stat finfo;

getcwd(cwd, PATH_MAX);

chdir(domain->getqout());
getcwd(dir, PATH_MAX);

if ((queue = opendir(dir)) == NULL)
	printf("Could not open queue directory\n");
else
	{
	struct dirent *file;
	while((file = readdir(queue)) != NULL)
		{
		FILE *header;
		if (*(file->d_name) == 'H')
			{
			if ((header = fopen(file->d_name, "r")) != NULL)
				{
				int first = 1;
				int fd;
				int locked = 1;
				char *dfile;
				struct flock flk;

				dfile = strdup(file->d_name);
				*dfile = 'R';				// relayed msg text
				if ((fd = open(dfile, O_RDONLY)) != -1)
					{
					fcntl(fd, F_GETLK, &flk);
					if (flk.l_type == F_UNLCK)
						locked = 0;
					close(fd);
					}
				stat(dfile, &finfo);
				printf("Message ID %s, size = %ld\n", file->d_name + 1, finfo.st_size);
				
				while(fgets(line, SMTP_MAX_LINE, header) != NULL)
					{
					switch(*line)
						{
						case 'F':
							printf("%c %18s : %s", locked ? '*' : ' ', "From", line + 1);
							break;
						case 'T':
							printf("%20s : %s", first?"To":"",line + 1);
							first = 0;
							break;
						case 'Q':
							{
							time_t t;
							t = strtol(line+1, NULL, 10);
							printf("%20s : %s", "Queue time", ctime(&t));
							}
							break;
						case 'R':
							printf("%20s : %s", "Last result", line + 1);
							break;
						case 'L':
							{
							time_t t, now;
							now = time(NULL);
							t = strtol(line+1, NULL, 10);
							printf("%20s : %s", "Last attempt", ctime(&t));
							printf("%20s : %s\n", "Retry", ((now - t) < retrytime) ? "Not expired" : "Ready to retry");
							}
							break;
						}
					}
				fclose(header);
				}
			}
		}
	closedir(queue);
	}
chdir(cwd);
}

int lookup(char *item, char *list[])
{
int n = 0;
while(list[n] != NULL)
	{
	if (strcmp(item, list[n]) == 0)
		return n;
	n++;
	}
return -1;
}

char *filename(char *path)
{
char *ptr;
if ((ptr = strrchr(path, '/')) == NULL)
	return path;
return ptr + 1;
}

//
// sendmail compatability mode (for queueing messages from stdin).
//
// Options:
// -F<from name>		Set the From full name
// -f<user>			Set the from user (root only)
// -i				Ignore dots on a line by themselves (normally terminates msg)
// -U				Initial (user) submission
// --				Trailing args are users

int sendmail(int argc, char *argv[])
{
int ignoredot = 0, users = 0, hasfrom = 0, usersubmit = 0, fullname = 0;
int n;
char *from = NULL;
char line[SMTP_MAX_LINE], tm[SMTP_MAX_LINE], ufullname[SMTP_MAX_LINE], ufrom[SMTP_MAX_LINE];
time_t t;
SMTP_SESSION *client;
STRING_ARRAY tolist;

printf("sendmail compatability mode\n");

if ((client = new SMTP_SESSION()) == NULL)
	return -1;

client->setdomain(domain);
client->setmode(SMTP_MODE_SERVER);
client->serverstatus(220);
client->process("HELO VorpalMail\r\n");

for(n=1;n<argc;n++)
	{
	if (!users && *argv[n] == '-')
		{
		switch(*(argv[n]+1))
			{
			case 'd':
				client->setdebug(strtol(argv[n]+2, NULL, 10));
				break;
			case 'i':
				ignoredot = 1;
				break;
			case 'U':
				usersubmit = 1;
				break;
			case '-':
				users = 1;
				break;
			case 'f':
				hasfrom = n;
				break;
			case 'F':
				fullname = n;
				break;
			}
		}
	else
		tolist.add(argv[n]);
	}

if (hasfrom && getuid() == 1)		// only root can be someone else
	strcpy(line, argv[hasfrom]+2);
else
	{
	struct passwd *pw;
	if ((pw = getpwuid(getuid())) == NULL)
		strcpy(line, "nobody");
	else
		strcpy(line, pw->pw_name);
	}
if (strchr(line, '@') == NULL)
	snprintf(ufrom, SMTP_MAX_LINE-1, "%s@%s", line, domain->getname());
else
	strcpy(ufrom, line);


if (fullname &&getuid() == 1)		// only root can be someone else
	strcpy(ufullname, argv[fullname]+2);
else
	{
	struct passwd *pw;
	if ((pw = getpwuid(getuid())) == NULL)
		strcpy(ufullname, "Unknown User");
	else
		strcpy(ufullname, pw->pw_gecos);
	if (*ufullname == 0)
		strcpy(ufullname, ufrom);
	}

snprintf(line, SMTP_MAX_LINE-1, "MAIL FROM:<%s>\r\n", ufrom);
client->process(line);

for(n=0;n<tolist.getsize();n++)
	{
	snprintf(line, SMTP_MAX_LINE-1, "RCPT TO:<%s>\r\n", tolist.get(n));
	client->process(line);
	}
client->process("DATA\r\n");

snprintf(line, SMTP_MAX_LINE-1, "From: \"%s\" <%s>\r\n", ufullname, ufrom);
client->process(line);

for(n=0;n<tolist.getsize();n++)
	{
	snprintf(line, SMTP_MAX_LINE-1, "%s %s%s\r\n", n==0 ? "To:" : "", tolist.get(n), n < (tolist.getsize()-1) ? "," : "");
	client->process(line);
	}

time(&t);
strftime(tm, SMTP_MAX_LINE-1, "%a, %d %b %Y %H:%M:%S %z", gmtime(&t));
snprintf(line, SMTP_MAX_LINE-1, "Date: %s\r\n", tm);
client->process(line);

while(fgets(line, SMTP_MAX_LINE-1, stdin) != NULL)
	{
	char *ptr;
	if ((ptr = strchr(line, '\n')) != NULL)
		*ptr = 0;
	if ((ptr = strchr(line, '\r')) != NULL)
		*ptr = 0;

	if (!ignoredot && strcmp(line, ".") == 0)
		break;
	
	if (*line == '.')
		snprintf(tm, SMTP_MAX_LINE-1, ".%s\r\n", line);
	else
		snprintf(tm, SMTP_MAX_LINE-1, "%s\r\n", line);

	client->process(tm);
	}
client->process(".\r\n");
client->process("QUIT\r\n");
return 0;
}

//
//
//
// Options:
//
// -l --list		show the queue contents
// -h --help		show a help message
// -d --daemon		run as a mail server daemon
// -q --queue		process the outgoing queue
//

int main(int argc, char *argv[])
{
int index = 0;
int val;
int mode = SMTP_MODE_NONE;

printf(PROGRAM " " VERSION "\n");

domain = new VIRTUAL_DOMAIN();
domain->load(CONFIG);

openlog(PROGRAM, LOG_CONS | LOG_PID, LOG_MAIL);

switch(lookup(filename(argv[0]), personality))
	{
	case PERS_MAILQ:
		listqueue();
		exit(0);
		break;
	case PERS_SENDMAIL:
	case PERS_SEND_MAIL:
		exit(sendmail(argc, argv));
		break;
	}

while((val = getopt_long(argc, argv, "r:lhdqD::", longoptions, &index)) != -1)
	{
	switch(val)
		{
		case 'h':
			help();
			exit(1);
			break;
		case 'l':
			listqueue();
			exit(0);
			break;
		case 'd':
			if (mode == SMTP_MODE_NONE)
				mode = SMTP_MODE_SERVER;
			else
				mode = -1;
			break;
		case 'q':
			if (mode == SMTP_MODE_NONE)
				mode = SMTP_MODE_CLIENT;
			else
				mode = -1;
			break;
		case 'D':
			if (optarg != NULL)
				debug = strtol(optarg, NULL, 10);
			else
				debug = 1;
			break;
		case 'm':
			domain->dump();
			exit(2);
		case 'r':
			{
			long n;
			if ((n = strtol(optarg, NULL, 10)) < 1 || n > 4320)
				printf("retry time must be in the range 1 - 4320 (minutes)\n");
			else
				retrytime = 60*n;
			}
			break;
		}
	index++;
	}

if (debug)
	printf("Debug mode: %d\n", debug);

switch(mode)
	{
	case SMTP_MODE_SERVER:
		exit(serverrun());
	case SMTP_MODE_CLIENT:
		exit(queuerun());
	}

printf("Must specify either --daemon or --queue mode.\n");
printf("%s --help for arguments.\n", argv[0]);

closelog();
}

