
//
//
// Project       : Vorpal-Mail
//
// Source        : session.cc
//
// Version       : 1.0
//
// Creator       : Blane Bramble                    Date          : 23rd October 2000
// Amended       : Blane Bramble                    Date          : 12th December 2000
//
// License       : Copyright Virtua-Web (GPL to be used once release 1.0 is achieved)
//
// Description   :
//
// Class implementation for SMTP session management. Provides a single SMTP session capable
// of transferring email compliant with RFC 821
//
// Changelog     :
//
// 9th Nov 2000
// Record locking code added - all files should be locked exclusive whilst in use (using fcntl).
// Session naming code altered.
//
// 13th Nov 2000
// Preliminary code added to fork() before delivering messages. If the fork fails then the
// message will be delivered in the main process.
//
// 22nd Nov 2000
// Added code to correctly cleanup received queue once a message has been delivered. Also
// initial code to call an external virus scanner.
//
// 5th December 2000
// Additional code to parse headers and automatically revert to body text if a header is not
// in the form "header: text" (where header is a single word). Needed by the command line
// sendmail compatability
//
// 7th December 2000
// Added support to reset (RSET) the connection at the start of each message,
// as if a RCPT TO failed, the next message was being injected into the
// aborted session.
//

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <syslog.h>
#include <errno.h>
#include <fcntl.h>

#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/wait.h>

#include <pwd.h>

#include "mail.h"
#include "array.h"
#include "session.h"
#include "virtual.h"
#include "smtp.h"

//
// Static initialisation:
//

char *SMTP_SESSION::commands[] = {
		"QUIT", 
		"HELP",
		"HELO",
		"EHLO",
		"MAIL",
		"RCPT",
		"DATA",
		"RSET",
		"NOOP",
		"VRFY",
		"EXPN",
		NULL };

char *SMTP_SESSION::command_help[] = {
	"Exit from " PROGRAM,
	"Display this help",
	"Request connection",
	"Request connection",
	"Start email from user",
	"Add user to recipient list",
	"Start transferring email data",
	"Reset session",
	"Do nothing",
	"Verify an email address",
	"Expand an email address",
	};

char *SMTP_SESSION::command_arg[] = {
	"",
	"",
	"<hostname>",
	"<hostname>",
	"<user@host>",
	"<user@host>",
	"<data>",
	"",
	"",
	"<user@host>",
	"<user@host>",
	};


char *SMTP_SESSION::hostname = NULL;

VIRTUAL_DOMAIN *SMTP_SESSION::domain = NULL;

long SMTP_SESSION::sequence = -1;
long SMTP_SESSION::pid = -1;
int SMTP_SESSION::debug = 0;

//
// Private class functions:
//

void SMTP_SESSION::initialise(void)
{

if (SMTP_SESSION::hostname == NULL)
	{
	char buf[128];
	gethostname(buf, 127);
	SMTP_SESSION::hostname = strdup(buf);
	}

if (SMTP_SESSION::sequence == -1)
	SMTP_SESSION::sequence = 1;

if (SMTP_SESSION::pid == -1)
	SMTP_SESSION::pid = getpid();

bodyfile = NULL;
bouncefile = NULL;
envelopefile = NULL;

sendbuffer = new LIST();
sendcount = 0;

bounce = 0; bouncing = 0;
filename = NULL;
line = NULL;
prev = next = NULL;
socket = -1;
mode = SMTP_MODE_NONE;
scratch = (char *)malloc(SMTP_MAX_LINE * 2);
spill = (char *)malloc(SMTP_MAX_LINE);
*spill = 0;
status = SMTP_STATUS_HELO;		// waiting for HELO
relayok = 0;				// by default relay is denied

from = NULL;

to = NULL;

dataheader = 1;
datamime = 0;
datamimesection = NULL;

ltime = qtime = 0;

*last_response = 0;
}

void SMTP_SESSION::cleanup(char *s)
{
if (s == NULL)
	s = filename;

if (status != SMTP_STATUS_COMPLETE)
	return;

if (this->mode == SMTP_MODE_SERVER)
	{
	unlink(makefilename(domain->getqin(), 'M', s));
	unlink(makefilename(domain->getqin(), 'E', s));
	unlink(makefilename(domain->getqin(), 'B', s));
	unlink(makefilename(domain->getqin(), 'H', s));
	}
else
	{
	unlink(makefilename(domain->getqout(), 'R', s));
	unlink(makefilename(domain->getqout(), 'F', s));
	unlink(makefilename(domain->getqout(), 'H', s));
	}
}

void SMTP_SESSION::closefiles(void)
{
if (bodyfile != NULL)
	{
	unlock(bodyfile);
	fclose(bodyfile);
	}
bodyfile = NULL;

if (bouncefile != NULL)
	{
	unlock(bouncefile);
	fclose(bouncefile);
	}
bouncefile = NULL;

if (envelopefile != NULL)
	{
	unlock(envelopefile);
	fclose(envelopefile);
	}
envelopefile = NULL;
}

void SMTP_SESSION::addheader(char *fmt, ...)
{
va_list argptr;
FILE *header;

va_start(argptr, fmt);
if ((header = fopen(makefilename(domain->getqin(), 'E', filename), "a")) != NULL)
	{
	vfprintf(header, fmt, argptr);
	fclose(header);
	}
va_end(argptr);
}

void SMTP_SESSION::virusmsg(char *log)
{
char *session, *msgto;
char tmp[SMTP_MAX_NAME];
LIST *local;

msgto = strdup(from);
session = filename;

closefiles();						// close old file
status = SMTP_STATUS_COMPLETE;

filename = strdup(sessionname());			// grab a new session ID
virushdr(session, log);

while(to->getsize() > 0)
	to->remove(0);

if (domain != NULL && (local = domain->expand(msgto)) != NULL)		// expand and convert to to user
	{
	LISTITEM *item;
	item = local->gethead();
	while(item != NULL)
		{
		addto(item->getkey());
		item = item->getnext();
		}
	delete local;
	}
else
	addto(msgto);
	
free(from);
snprintf(tmp, SMTP_MAX_NAME-1, "postmaster@%s", domain->getname());
from = strdup(tmp);

localdeliver();
writeheader();
remotedeliver();

cleanup(filename);

// TODO: lots!

free(msgto);
free(filename);
filename = session;
}

int SMTP_SESSION::scanvirus(char *log)
{
char line[MAX_CMD];
char tmpname[] = "/tmp/vmailtmpXXXXXX";
int fd, pid;
if ((fd = mkstemp(tmpname)) == -1)
	syslog(LOG_MAIL + LOG_DEBUG, "** Could not make temporary file, virus scan aborted");
else
	{
	FILE *out;

	snprintf(log, 30, "%s.log", tmpname);

	if ((out = fdopen(fd, "w")) == NULL)
		{
		syslog(LOG_MAIL + LOG_DEBUG, "** Could not open temporary file, virus scan aborted");
		close(fd);
		}
	else
		{
		FILE *in;
		int status;

		if ((in = fopen(makefilename(domain->getqin(),'E',filename), "r")) == NULL)
			{
			syslog(LOG_MAIL + LOG_DEBUG, "** Could not open envelope file, virus scan aborted");
			fclose(out);
			return 0;
			}

		lock(in);	
		while(fgets(line, MAX_CMD-1, in) != NULL)
			fputs(line, out);
		unlock(in);
		fclose(in);

		fputs("\n", out);		// blank line seperates

		if ((in = fopen(makefilename(domain->getqin(),'M',filename), "r")) == NULL)
			{
			syslog(LOG_MAIL + LOG_DEBUG, "** Could not open message file, virus scan aborted");
			fclose(out);
			return 0;
			}
		
		lock(in);	
		while(fgets(line, MAX_CMD-1, in) != NULL)
			fputs(line, out);
		unlock(in);
		fclose(in);

		fclose(out);

		if (debug)
			syslog(LOG_MAIL + LOG_DEBUG, "Virus scanner started (tmp file %s, session %s)", tmpname, filename);

		switch((pid = fork()))
			{
			case -1:
				syslog(LOG_MAIL + LOG_DEBUG, "** Could not fork, virus scan aborted");
				break;
			case 0:		// child!
				if (debug)
					syslog(LOG_MAIL + LOG_DEBUG, "Exec: %s %s", domain->getscanner(), tmpname+5);
				execl(domain->getscanner(), domain->getscanner(), tmpname + 5, log, NULL);
				syslog(LOG_MAIL + LOG_DEBUG, "Exec failed, %d (%s)", errno, strerror(errno));
				_exit(1);
				break;
			default:
				if (waitpid(pid, &status, 0) != pid)
					syslog(LOG_MAIL + LOG_DEBUG, "** Error waiting for scanner to exit, virus scan aborted");
				else
					{
					if (debug)
						syslog(LOG_MAIL + LOG_DEBUG, "Virus scanner finished");
					if ((in = fopen(tmpname, "r")) == NULL)		// file inaccessable!
						return 1;
					else
						{
						fclose(in);
						unlink(tmpname);		// delete temporary file
						addheader("X-Vorpal: Virus Scanner " PROGRAM "\n");
						return 0;
						}
					}
				break;
			}
		}
	}
return 0;
}

void SMTP_SESSION::reset(void)
{
int i;
char *ptr;

bounce = 0; 
relayok = 0; 
*spill = 0;
dataheader = 1;
datamime = 0;

closefiles();
cleanup(NULL);

if (filename != NULL)
	free(filename);
filename = NULL;

line = NULL;

status = SMTP_STATUS_HELO;

if (from != NULL)
	free(from);
from = NULL;

if (datamimesection != NULL)
	free(datamimesection);

if (to != NULL)
	delete to;
to = NULL;
}

void SMTP_SESSION::closesocket(void)
{
reset();
if (socket != -1)
	{
	shutdown(socket, 2);
	close(socket);
	}
socket = -1;
}

int SMTP_SESSION::lookup(char *s)
{
int n=0;
while(SMTP_SESSION::commands[n] != NULL)
	{
	if (strcasecmp(SMTP_SESSION::commands[n], s) == 0)
		return n;
	n++;
	}
return -1;
}

int SMTP_SESSION::splitheader(char *line, char *h, char *t)
{
char *from, *to;
if (strchr(line, ':') == NULL)
	return 0;

from = line;
to = h;

while(isspace(*from))
	from++;

while(*from != 0 && *from != ':')
	*to++ = *from++;
*to = 0;

if (*from == ':')
	from++;
while(isspace(*from))
	from++;

to = t;
while(*from != 0)
	*to++ = *from++;
*to = 0;
if (*t != 0)
	return 2;
return 1;
}

void SMTP_SESSION::process_data(char *data)
{
if (SMTP_SESSION::debug > 2)
	syslog(LOG_MAIL + LOG_DEBUG, "<< %s\n", data);

status = SMTP_STATUS_DATA;

if (strcmp(data, ".") == 0)
	{
	int pid;
	char virlog[40];

	if (qtime == 0)
		qtime = time(NULL);
	unlock(bodyfile);
	fclose(bodyfile);
 	bodyfile = NULL;
       	serverstatus(250);
	switch((pid = fork()))
		{

		case -1:
			syslog(LOG_MAIL + LOG_DEBUG, "** Could not fork. Delivering %s in same process", filename);
		case 0:
			if (scanvirus(virlog))
				{
				syslog(LOG_MAIL + LOG_DEBUG, "** Virus found in message %s", filename);
				closefiles();
				virusmsg(virlog);
				unlink(virlog);
				cleanup(NULL);
				}
			else
				{
				unlink(virlog);
				localdeliver();
				writeheader();
				remotedeliver();

				if (bounce)
					sendbounce();

				cleanup(NULL);
				}
			if (pid == 0)
				_exit(0);
			break;
		default:
			if (debug)
				syslog(LOG_MAIL + LOG_DEBUG, "Child %d is delivering mail", pid);
			break;
		}
	status = SMTP_STATUS_MAIL;
	}
else
	{
	if (dataheader)
		{
		char head[SMTP_MAX_LINE], tail[SMTP_MAX_LINE];
		if (!isspace(*data) && (!splitheader(data, head, tail) || strchr(head, ' ')!= NULL))		// not a header!
			{
			unlock(envelopefile);
			fclose(envelopefile);
			envelopefile = NULL;
			dataheader = 0;
			if (*data != 0)
				fprintf(bodyfile, "%s\n", data);
			}
		else
			{
			if (strcasecmp(data, "MIME-Version: 1.0") == 0)
				datamime = 1;
			if (envelopefile != NULL)
				fprintf(envelopefile, "%s\n", data);
			else
				{
				addbounce("Internal error: could not spool envelope.\n");
				serverstatus(451);
				sendbounce();
				status = SMTP_STATUS_MAIL;
				return;
				}
			}
		}
	else
		{
		if (bodyfile != NULL)
			{
			if (*data == '.')
				fprintf(bodyfile, "%s\n", data+1);
			else
				fprintf(bodyfile, "%s\n", data);
			}
		else
			{
			addbounce("Internal error: could not spool message.\n");
			serverstatus(451);
			sendbounce();
			status = SMTP_STATUS_MAIL;
			return;
			}
		}
	
	}
}

int SMTP_SESSION::parsename(char *from, char *pre, char *to)
{
while(*pre != 0)		// check the correct prefix is present
	{
	if (toupper(*from) != toupper(*pre))
		return 0;
	from++; pre++;
	}

while(isspace(*from) && *from != 0)
	from++;

if (*from++ != '<')		// not in angle brackets
	return 0;

while(*from != '>')
	*to++ = *from++;
*to = 0;
return 1;
}

int SMTP_SESSION::localuser(char *msgto, VIRTUAL_DOMAIN *dom)
{
int statuslocal = 1, statusrelay = -1, statusinvalid = 0;		// possible results for local user, remote user, local but doesnt exist

if (dom != NULL)
	{
	switch(dom->getunknown())
		{
		case VIRTDOM_UNKNOWN_ACCEPT:							// invalid treated as local
			statusinvalid = 1;
			break;
		case VIRTDOM_UNKNOWN_REJECT:							// default settings
			break;
		case VIRTDOM_UNKNOWN_RELAY:							// invalid treated as remote
			statusinvalid = -1;
			break;
		case VIRTDOM_UNKNOWN_MAILTO:							// invalid treates as alias
			if (getpwnam(dom->getunknownrelay()) != NULL)				// mailto is a local user
				statusinvalid = 2;						// so replace any unknown users
			else
				statusinvalid = 0;						// otherwise reject unknown
			break;
		}
	if (getpwnam(msgto) == NULL)
		return statusinvalid;
	else
		return statuslocal;
	}
else
	return statusrelay;										// remote user
}


void SMTP_SESSION::sendbounce(void)
{
char *session, *msgto;
char tmp[SMTP_MAX_NAME];
LIST *local;

bouncing = 1; bounce = 0;				// stop bounce msg from being bounced internally

msgto = strdup(from);
session = filename;

closefiles();						// close old file
status = SMTP_STATUS_COMPLETE;

filename = strdup(sessionname());			// grab a new session ID
errormail(session);					// create a new email

while(to->getsize() > 0)
	to->remove(0);

if (domain != NULL && (local = domain->expand(msgto)) != NULL)		// expand and convert to to user
	{
	LISTITEM *item;
	item = local->gethead();
	while(item != NULL)
		{
		addto(item->getkey());
		item = item->getnext();
		}
	delete local;
	}
else
	addto(msgto);
	
free(from);
snprintf(tmp, SMTP_MAX_NAME-1, "postmaster@%s", domain->getname());
from = strdup(tmp);

localdeliver();
writeheader();
remotedeliver();

cleanup(filename);

// TODO: lots!

free(msgto);
free(filename);
filename = session;
}

int SMTP_SESSION::lock(FILE *fp)
{
struct flock flk;

flk.l_type = F_WRLCK;
flk.l_whence = 0;
flk.l_start = 0;
flk.l_len = 0;

if (fcntl(fileno(fp), F_SETLK, &flk) == -1)
	return 0;
return 1;
}

void SMTP_SESSION::unlock(FILE *fp)
{
struct flock flk;

flk.l_type = F_UNLCK;
flk.l_whence = 0;
flk.l_start = 0;
flk.l_len = 0;

fcntl(fileno(fp), F_SETLK, &flk);
}

void SMTP_SESSION::virushdr(char *sess, char *log)
{
FILE *msgfile, *data;
char tm[SMTP_MAX_LINE];
time_t t;

time(&t);
strftime(tm, sizeof(tm), "%a, %d %b %Y %H:%M:%S %z", gmtime(&t));

if ((msgfile = fopen(makefilename(domain->getqin(), 'E', filename), "w")) == NULL)
	return;

lock(msgfile);

fprintf(msgfile, "Date: %s\n", tm);
fprintf(msgfile, "From: postmaster@%s\n", domain->getname());
fprintf(msgfile, "To: %s\n", from);
fprintf(msgfile, "Subject: ** Virus Warning **\n");

unlock(msgfile);
fclose(msgfile);

if ((msgfile = fopen(makefilename(domain->getqin(), 'M', filename), "w")) == NULL)
	return;

lock(msgfile);	

fprintf(msgfile, "This is an automatically generated message. Do not reply to it.\n\n");
fprintf(msgfile, "A message sent by you has been received and was found to contain a virus.\n\n");
fprintf(msgfile, "As a result this message has not been delivered to it's recipient(s).\n\n");
fprintf(msgfile, "The result of the virus scan, and your original message headers are\n\n");
fprintf(msgfile, "reproduced below for your information.\n\n");

fprintf(msgfile, "------ Scanner transscript ------\n\n");

if ((data = fopen(log, "r")) == NULL)
	fprintf(msgfile, "    [ ERROR: Transcript lost ]\n");
else
	{
	
	lock(data);
	
	while(fgets(tm, SMTP_MAX_LINE-1, data)!= NULL)
		fprintf(msgfile, "    %s", tm);
		
	unlock(data);
		
	fclose(data);
	}

fprintf(msgfile, "\n------ Message Headers ------\n\n");

if ((data = fopen(makefilename(domain->getqin(), 'E', sess), "r")) == NULL)
	fprintf(msgfile, "    [ ERROR: Message Headers lost ]\n");
else
	{
	
	lock(data);
	
	while(fgets(tm, SMTP_MAX_LINE-1, data)!= NULL)
		fprintf(msgfile, "    %s", tm);
		
	unlock(data);		
		
	fclose(data);
	}

unlock(msgfile);	
	
fclose(msgfile);
}


void SMTP_SESSION::errormail(char *sess)
{
FILE *msgfile, *data;
char *ptr, tm[SMTP_MAX_LINE];
time_t t;
VIRTUAL_DOMAIN *dom;

time(&t);
strftime(tm, sizeof(tm), "%a, %d %b %Y %H:%M:%S %z", gmtime(&t));

if ((msgfile = fopen(makefilename(domain->getqin(), 'E', filename), "w")) == NULL)
	return;

lock(msgfile);	

//
// Firstly create a simple envelope file
//

fprintf(msgfile, "Date: %s\n", tm);
fprintf(msgfile, "From: postmaster@%s\n", domain->getname());
fprintf(msgfile, "To: %s\n", from);
fprintf(msgfile, "Subject: Error unable to deliver message\n");

unlock(msgfile);

fclose(msgfile);

//
// Now create the message body
//

if ((msgfile = fopen(makefilename(domain->getqin(), 'M', filename), "w")) == NULL)
	return;

lock(msgfile);	
	
fprintf(msgfile, "An error occurred whilst trying to deliver a message from you.\n\n");
fprintf(msgfile, "As a result this message has not been delivered to all the recipients.\n\n");

fprintf(msgfile, "------ Error transscript ------\n\n");

if ((data = fopen(makefilename(domain->getqin(), 'B', sess), "r")) == NULL)
	fprintf(msgfile, "    [ ERROR: Transcript lost ]\n");
else
	{
	
	lock(data);
	
	while(fgets(tm, SMTP_MAX_LINE-1, data)!= NULL)
		fprintf(msgfile, "    %s", tm);
		
	unlock(data);
		
	fclose(data);
	}

fprintf(msgfile, "\n------ Message Headers ------\n\n");

if ((data = fopen(makefilename(domain->getqin(), 'E', sess), "r")) == NULL)
	fprintf(msgfile, "    [ ERROR: Message Headers lost ]\n");
else
	{
	
	lock(data);
	
	while(fgets(tm, SMTP_MAX_LINE-1, data)!= NULL)
		fprintf(msgfile, "    %s", tm);
		
	unlock(data);		
		
	fclose(data);
	}

unlock(msgfile);	
	
fclose(msgfile);
}

int SMTP_SESSION::dolocaldeliver(char *to, char *sess, VIRTUAL_DOMAIN *dom)
{
char *cmdline, *ptr;
char tmpcmd[MAX_CMD];
FILE *data, *proc;
int pstat;

ptr = tmpcmd;
if ((cmdline = dom->getlocal()) == NULL)
	return 1;

while(*cmdline != 0)
	{
	if (*cmdline == '%')
		{
		switch(*(++cmdline))
			{
			default:
				break;
			case 'h':
				strcpy(ptr, domain->getname());
				ptr += strlen(ptr);
				cmdline++;
				break;
			case 'u':
				strcpy(ptr, to);
				ptr += strlen(ptr);
				cmdline++;
				break;
			case 'f':
				strcpy(ptr, from);
				ptr += strlen(ptr);
				cmdline++;
				break;
			}
		}
	else
		*ptr++ = *cmdline++;
	}
*ptr = 0;

if (debug)
	syslog(LOG_MAIL + LOG_DEBUG, "Local delivery with %s", tmpcmd);

fflush(NULL);

if ((proc = popen(tmpcmd, "w")) == NULL)			// TODO: need to add error checking
	{
	syslog(LOG_MAIL + LOG_DEBUG, "** Could not run local delivery program");
	return 1;
	}

if ((data = fopen(makefilename(domain->getqin(),'E',sess), "r")) == NULL)
	{
	syslog(LOG_MAIL + LOG_DEBUG, "** Could not open envelope for local delivery");
	return 1;
	}
	
lock(data);	
	
while(fgets(tmpcmd, MAX_CMD-1, data) != NULL)
	fputs(tmpcmd, proc);

unlock(data);

fclose(data);

fputs("\n", proc);		// add blank line to seperate headers

if ((data = fopen(makefilename(domain->getqin(),'M',sess), "r")) == NULL)
	{
	syslog(LOG_MAIL + LOG_DEBUG, "** Could not open message body for local delivery");
	return 1;
	}
lock(data);	
	
while(fgets(tmpcmd, MAX_CMD-1, data) != NULL)
	fputs(tmpcmd, proc);

unlock(data);

fclose(data);

if (WIFEXITED((pstat = pclose(proc))))
	return WEXITSTATUS(pstat);
return -1;
}

void SMTP_SESSION::localdeliver(void)
{
static char tmp[SMTP_MAX_NAME];
int n = 0;
char *msgto, *ptr;
VIRTUAL_DOMAIN *dom;

while((msgto = to->get(n)) != NULL)
	{
	strcpy(tmp, msgto);
	msgto = tmp;

	ptr = strchr(msgto, '@');
	if (ptr == NULL || (domain != NULL && (dom = domain->finddomain(ptr+1)) != NULL))
		{
		if (ptr != NULL)
			*ptr = 0;

		switch(localuser(msgto, dom))
			{
			case 2:				// mailto local
				msgto = dom->getunknownrelay();
			case 1:				// local user
				if (dolocaldeliver(msgto, filename, dom) != 0)
					addbounce("<%s@%s> unable to deliver local mail.\n", msgto, dom->getname());
				to->remove(n);
				break;
			case 0:
				addbounce("<%s@%s> user unknown.\n", msgto, dom->getname());
				to->remove(n);
				break;
			case -2:			// alias (shouldn't occur at this stage!)
			case -1:			// remote user
				n++;
			break;
			}
		}
	else
		n++;					// remote user
	}
}

void SMTP_SESSION::remotedeliver(void)
{
if (to->getsize() != 0)
	doremotedeliver();
}

void SMTP_SESSION::doremotedeliver(void)
{
FILE *fpin, *fpout;
char buf[512];
int n;

fpin = fopen(makefilename(domain->getqin(), 'M', filename), "r");
fpout = fopen(makefilename(domain->getqout(), 'R', filename), "w");

if (fpin == NULL || fpout == NULL)		// TODO: need to trap this error and report it!
	return;

lock(fpin); lock(fpout);	
	
while((n = fread(buf,  1, 512, fpin)) > 0)
	fwrite(buf, 1, n, fpout);

// TODO: need to check feof() rather than ferror() has caused fread to finish	
		
unlock(fpin); unlock(fpout);

fclose(fpin);
fclose(fpout);

fpin = fopen(makefilename(domain->getqin(), 'E', filename), "r");
fpout = fopen(makefilename(domain->getqout(), 'F', filename), "w");

if (fpin == NULL || fpout == NULL)		// TODO: need to trap this error and report it!
	return;

lock(fpin);
lock(fpout);	
	
while((n = fread(buf,  1, 512, fpin)) > 0)
	fwrite(buf, 1, n, fpout);

// TODO: need to check feof() rather than ferror() has caused fread to finish	

unlock(fpin);
unlock(fpout);
		
fclose(fpin);
fclose(fpout);
}

void SMTP_SESSION::addreceive(void)
{
long l;
int n, quad[4];
time_t t;
char tm[128];
struct hostent *host;

time(&t);
strftime(tm, sizeof(tm), "%a, %d %b %Y %H:%M:%S %z", gmtime(&t));

l = ntohl(saddr.sin_addr.s_addr);
for(n=0;n<4;n++)
	{
	quad[n] = l & 0xff;
	l >>= 8;
	}

host = gethostbyaddr((char *)&saddr.sin_addr.s_addr, sizeof(saddr.sin_addr.s_addr), AF_INET);

fprintf(envelopefile, "Received: from %s (%s [%d.%d.%d.%d])\n", host->h_name, host->h_name, quad[3],quad[2], quad[1], quad[0]);
fprintf(envelopefile, "\tby %s (" PROGRAM " " VERSION ") with SMTP id %s;\n", SMTP_SESSION::hostname, filename);
fprintf(envelopefile, "\t%s\n", tm);
}

void SMTP_SESSION::process_client(char *data)
{
int code;

code = *data;
if (SMTP_SESSION::debug > 2 || (SMTP_SESSION::debug > 1 && this->status != SMTP_STATUS_DATASEND))
	syslog(LOG_MAIL + LOG_DEBUG, "<< %s\n", data);

strcpy(last_response, data);

switch(this->status)
	{
	case SMTP_STATUS_OPEN:
		switch(code)
			{
			case '2':
				clientstatus(SMTP_STATUS_HELO);
				break;
			default:
				this->status = SMTP_STATUS_ABORT;
				break;
			}
		break;
	case SMTP_STATUS_RSET:
		switch(code)
			{
			case '2':
				clientstatus(SMTP_STATUS_MAIL, this->from);
				break;
			default:
				this->status = SMTP_STATUS_ABORT;
				break;
			}
		break;
	case SMTP_STATUS_HELO:
		switch(code)
			{
			case '2':
				clientstatus(SMTP_STATUS_MAIL, this->from);
				break;
			default:
				this->status = SMTP_STATUS_ABORT;
				break;
			}
		break;
	case SMTP_STATUS_MAIL:
		switch(code)
			{
			case '2':
				if (this->to->get(0) == NULL)
					this->status = SMTP_STATUS_ABORT;
				else
					{
					this->loop = 1;
					clientstatus(SMTP_STATUS_RCPT, this->to->get(0));
					}
				break;
			default:
				this->status = SMTP_STATUS_ABORT;
			}
		break;
	case SMTP_STATUS_RCPT:
		switch(code)
			{
			case '2':
				if (this->to->get(this->loop) != NULL)
					{
					clientstatus(SMTP_STATUS_RCPT, this->to->get(this->loop));
					this->loop++;
					}
				else
					clientstatus(SMTP_STATUS_DATA);
				break;
			default:
				this->status = SMTP_STATUS_ABORT;
			}
		break;
	case SMTP_STATUS_DATA:
		switch(code)
			{
			case '3':
				if (this->envelopefile == NULL || this->bodyfile == NULL)
					this->status = SMTP_STATUS_ABORT;
				else
					this->status = SMTP_STATUS_DATASEND;
				break;
			default:
				this->status = SMTP_STATUS_ABORT;
				break;
			}
		break;
	case SMTP_STATUS_DATASEND:		// need to move logic to send a line at a time and return to here
	case SMTP_STATUS_SENDRETRY:
		if (this->envelopefile != NULL)
			{
			if (status == SMTP_STATUS_SENDRETRY || fgets(spill, SMTP_MAX_LINE, this->envelopefile) != NULL)
				{
				char *ptr;
				if ((ptr = strchr(spill, '\n')) != NULL)
					*ptr = 0;
				if ((ptr = strchr(spill, '\r')) != NULL)
					*ptr = 0;
				clientstatus(SMTP_STATUS_DATASEND, spill);
				}
			else
				{
				unlock(envelopefile);
				fclose(this->envelopefile);
				this->envelopefile = NULL;
				}
			}
		else if (this->bodyfile != NULL)
			{
			if (status == SMTP_STATUS_SENDRETRY || fgets(spill, SMTP_MAX_LINE, this->bodyfile) != NULL)
				{
				char *ptr;
				if ((ptr = strchr(spill, '\n')) != NULL)
					*ptr = 0;
				if ((ptr = strchr(spill, '\r')) != NULL)
					*ptr = 0;
				clientstatus(SMTP_STATUS_DATASEND, spill);
				}
			else
				{
				clientstatus(SMTP_STATUS_DATASEND, NULL);
				unlock(bodyfile);
				fclose(this->bodyfile);
				this->bodyfile = NULL;
				this->status = SMTP_STATUS_SENT;
				}
			}
		break;
	case SMTP_STATUS_DATATEXT:
		clientstatus(SMTP_STATUS_DATASEND, NULL);
		break;
	case SMTP_STATUS_SENT:
		switch(code)
			{
			case '2':
				this->status = SMTP_STATUS_COMPLETE;
				cleanup(NULL);			// remove files
				break;
			default:
				this->status = SMTP_STATUS_ABORT;
				break;
			}
		break;
	}

if (status == SMTP_STATUS_ABORT || status == SMTP_STATUS_COMPLETE)
	{
	ltime = time(NULL);
	if (status == SMTP_STATUS_ABORT)
		writeheader();
	}
}

char *SMTP_SESSION::sessionname(void)
{
static char namebuf[PATH_MAX];
time_t ut;

time(&ut);

// session name is unixtime shifted right by 12 bits and then the low 16 bits masked (which correspond to bits 12-28 of the original)
// this main section will thus change every 68 minutes, wrapping every 3106 days (by which time any stuck message should have bounced, so
// the time:sequence pairing should never be duplicated).

snprintf(namebuf, PATH_MAX-1, "%.4x%.4x%.6x", SMTP_SESSION::pid, (unsigned int)((ut>>12) & 0xffff), (int)SMTP_SESSION::sequence++);

return namebuf;
}

void SMTP_SESSION::addbounce(char *fmt, ...)
{
va_list argptr;

if (bouncing)
	return;

va_start(argptr, fmt);
vfprintf(bouncefile, fmt, argptr);
bounce++;
va_end(argptr);
}

//
// PUBLIC class functions:
//

SMTP_SESSION::SMTP_SESSION(void)
{
initialise();
// Need to make this point to the local machine by default
saddr.sin_family = AF_INET;
saddr.sin_addr.s_addr = htonl(0x7f000001);
saddr.sin_port = htons(-1);
}

SMTP_SESSION::SMTP_SESSION(int s, struct sockaddr_in *sa)
{
initialise();
memcpy(&this->saddr, sa, sizeof(struct sockaddr_in));
socket = s;
}

SMTP_SESSION::~SMTP_SESSION()
{
free(spill);
free(scratch);
closesocket();			// does reset etc.
}

void SMTP_SESSION::setnext(SMTP_SESSION *n)
{
this->next = n;
n->prev = this;
}

int SMTP_SESSION::setmode(SMTP_MODE m)
{
if (this->mode != SMTP_MODE_NONE)
	return 0;
if ((this->mode = m) == SMTP_MODE_CLIENT)
	this->status = SMTP_STATUS_OPEN;
return 1;
}

//
// SMTP_SESSION forms a file name by combining a queue directory, a file type and the session name.
//
// file types are:
//
// 'M'		Message body received in queuein
// 'R'		Message body relayed to queueout
// 'H'		SMTP_SESSION header information
// 'E'		RFC 822 header information
// 'F'		RFC 822 header relayed to queueout
// 'B'		Bounce information
//

char *SMTP_SESSION::makefilename(char *queue, char type, char *file)
{
static char path[PATH_MAX];

snprintf(path, PATH_MAX-1, "%s/%c%s", queue, type, file);

return path;
}

int SMTP_SESSION::openfiles(char *fn)
{
if (filename == NULL)
	filename = strdup(fn);
	
if (this->mode == SMTP_MODE_SERVER)
	{
	if (bodyfile == NULL)
		bodyfile = fopen(makefilename(domain->getqin(), 'M', fn), "w");	

	if (bouncefile == NULL)
		bouncefile = fopen(makefilename(domain->getqin(), 'B', fn), "w");

	if (envelopefile == NULL)
		envelopefile = fopen(makefilename(domain->getqin(), 'E', fn), "w");


	if (bodyfile == NULL || bouncefile == NULL || envelopefile == NULL)
		return 0;
		
	lock(bodyfile);
	lock(bouncefile);
	lock(envelopefile);
		
	}
else if (this->mode == SMTP_MODE_CLIENT)
	{
	if (bodyfile == NULL)
		bodyfile = fopen(makefilename(domain->getqout(), 'R', fn), "r");	

	if (envelopefile == NULL)
		envelopefile = fopen(makefilename(domain->getqout(), 'F', fn), "r");

	if (bodyfile == NULL || envelopefile == NULL)
		return 0;
		
	lock(bodyfile);
	lock(envelopefile);
	}
return 1;
}


void SMTP_SESSION::setfrom(char *f)
{
if (from == NULL)
	{
	from = strdup(f);
	relayok = checkrelay(from);
	}
}

void SMTP_SESSION::addto(char *t)
{
if (this->to == NULL)
	this->to = new STRING_ARRAY();
	
if (this->to->find(t) == -1)		// only add if it doesnt exist already
	this->to->add(t);
}

void SMTP_SESSION::writeheader(void)
{
FILE *hfp;
int n = 0;

if (to->getsize() == 0)
	{
	if (debug > 1)
		syslog(LOG_MAIL + LOG_INFO, "** No remote users to write header for.\n");
	return;
	}

hfp = fopen(makefilename(domain->getqout(), 'H', filename), "w");
lock(hfp);
fprintf(hfp, "F%s\n", from);
if (this->to != NULL)
	{
	while(this->to->get(n) != NULL)
		{
		fprintf(hfp, "T%s\n", this->to->get(n));
		n++;
		}
	}

if (this->qtime != 0)
	fprintf(hfp, "Q%ld\n", qtime);

if (this->ltime != 0)
	fprintf(hfp, "L%ld\n", ltime);
if (*(this->last_response) != 0)
	fprintf(hfp, "R%s\n", this->last_response);

unlock(hfp);
fclose(hfp);
}

void SMTP_SESSION::process(char *data)
{
char *from,*to,cmd[SMTP_MAX_LINE];

from = data;
strcpy(cmd, spill);

to = (cmd + strlen(cmd));

while(*from != 0)
	{
	if (*from == '\r')
		{
		*to = 0;
		switch(this->mode)
			{
			case SMTP_MODE_SERVER:
				process_server(cmd);
				break;
			case SMTP_MODE_CLIENT:
				process_client(cmd);
				break;
			}
		to = cmd;
		if (*(from+1) == '\n')
			from++;
		}
	else
		*to++ = *from;
	from++;
	}
*to = 0;
strcpy(spill, cmd);		// buffer partial command back
}


int SMTP_SESSION::sendf(char *fmt, ...)
{
char tmp[1024], *ptr;
va_list argptr;
va_start(argptr, fmt);
vsnprintf(tmp, 1023, fmt, argptr);

if ((ptr = strdup(tmp)) == NULL)
	return 0;

if (sendbuffer->add(NULL, ptr) == NULL)
	{
	free(ptr);
	return 0;
	}

if ((SMTP_SESSION::debug > 2) || (SMTP_SESSION::debug > 1 && this->status != SMTP_STATUS_DATASEND))
	syslog(LOG_MAIL + LOG_DEBUG, ">> %s\n", tmp);

va_end(argptr);

return 1;
}

int SMTP_SESSION::pending(void)
{
if (sendbuffer->gethead() == NULL)
	return 0;
return 1;
}

void SMTP_SESSION::dosend(void)
{
if (pending())
	{
	LISTITEM *item;
	if ((item = sendbuffer->gethead()) != NULL)
		{
		char *ptr;
		int tosend;

		ptr = (char *)item->getdata();
		tosend = strlen(ptr) - sendcount;

		if ((sendcount += send(socket, ptr + sendcount, tosend, 0)) == tosend)
			{
			free(sendbuffer->del(item));			// when we delete the list item, free the string as well
			sendcount = 0;
			}
		}
	}
}

void SMTP_SESSION::clientstatus(SMTP_STATUS stat)
{
clientstatus(stat, "[none]");
}

void SMTP_SESSION::clientstatus(SMTP_STATUS stat, char *str)
{
switch(stat)
	{
	case SMTP_STATUS_TIMEOUT:
		this->status = SMTP_STATUS_ABORT;
		ltime = time(NULL);
		writeheader();
		break;
	case SMTP_STATUS_HELO:
		if (sendf("HELO %s\r\n", hostname))
			this->status = SMTP_STATUS_HELO;
		else
			this->status = SMTP_STATUS_ABORT;
		break;
	case SMTP_STATUS_MAIL:
		if (sendf("MAIL FROM:<%s>\r\n", str))
			this->status = SMTP_STATUS_MAIL;
		else
			this->status = SMTP_STATUS_ABORT;
		break;
	case SMTP_STATUS_RCPT:
		if (sendf("RCPT TO:<%s>\r\n", str))
			this->status = SMTP_STATUS_RCPT;
		else
			this->status = SMTP_STATUS_ABORT;
		break;
	case SMTP_STATUS_DATA:
		if (sendf("DATA\r\n"))
			this->status = SMTP_STATUS_DATA;
		else
			this->status = SMTP_STATUS_ABORT;
		break;
	case SMTP_STATUS_DATASEND:
		this->status = SMTP_STATUS_DATASEND;
		if (str == NULL)
			{
			if (sendf(".\r\n"))
				this->status = SMTP_STATUS_SENT;
			else
				this->status = SMTP_STATUS_SENDRETRY;
			}
		else if (*str == '.')
			{
			if (sendf(".%s\r\n", str))
				this->status = SMTP_STATUS_DATASEND;
			else
				this->status = SMTP_STATUS_SENDRETRY;
			}
		else
			{
			if (sendf("%s\r\n", str))
				this->status = SMTP_STATUS_DATASEND;
			else
				this->status = SMTP_STATUS_SENDRETRY;
			}
		break;
	}
}

void SMTP_SESSION::setdomain(VIRTUAL_DOMAIN *d)
{
domain = d;
}

void SMTP_SESSION::sendreset(void)
{
if (mode != SMTP_MODE_CLIENT)
	return;
if (sendf("RSET\r\n"))
	this->status = SMTP_STATUS_RSET;
else
	this->status = SMTP_STATUS_ABORT;
}

void SMTP_SESSION::sendquit(void)
{
if (mode != SMTP_MODE_CLIENT)
	return;
if (sendf("QUIT\r\n"))
	this->status = SMTP_STATUS_COMPLETE;
else
	this->status = SMTP_STATUS_ABORT;
}

void SMTP_SESSION::pulse(void)
{
if (timeout)
	{
	if (--timeout == 0)
		{
		switch(mode)
			{
			case SMTP_MODE_SERVER:
				serverstatus(SMTP_STATUS_TIMEOUT);
				break;
			case SMTP_MODE_CLIENT:
				clientstatus(SMTP_STATUS_TIMEOUT);
				break;
			}
		}
	}
}
